/*
 * Copyright (C) 2006-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file user_list.c user handling for para_server */

#include "para.h"
#include "error.h"
#include "fd.h"
#include "string.h"
#include "user_list.h"

static struct list_head user_list;

/*
 * lookup user in user list file. Fills in a user struct containing
 * filename of the user's public key as well as the permissions of that user.
 * Returns 1 on success, 0 if user does not exist and < 0 on errors.
 */
static void populate_user_list(char *user_list_file)
{
	FILE *file_ptr = NULL;
	char *char_ptr;
	char line[MAXLINE];
	/* keyword, user, key, perms */
	char w[MAXLINE], n[MAXLINE], k[MAXLINE], p[MAXLINE], tmp[4][MAXLINE];
	int num, ret;

	file_ptr = fopen(user_list_file, "r");
	ret = -E_USERLIST;
	if (!file_ptr)
		goto out;
	for (;;) {
		struct user *u;
		ret = para_fgets(line, MAXLINE, file_ptr);
		if (ret < 0)
			PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
		if (ret <= 0)
			break;
		if (sscanf(line,"%200s %200s %200s %200s", w, n, k, p) < 3)
			continue;
		if (strcmp(w, "user"))
			continue;
		PARA_DEBUG_LOG("found entry for %s\n", n);
		u = para_malloc(sizeof(struct user));
		u->name = para_strdup(n);
		ret = get_rsa_key(k, &u->rsa, LOAD_PUBLIC_KEY);
		if (ret < 0)
			break;
		char_ptr = p;
		num = sscanf(char_ptr, "%200[A-Z_],%200[A-Z_],%200[A-Z_],%200[A-Z_]",
			tmp[0], tmp[1], tmp[2], tmp[3]);
		PARA_DEBUG_LOG("found %i perm entries\n", num);
		u->perms = 0;
		while (num > 0) {
			num--;
			if (!strcmp(tmp[num], "VSS_READ"))
				u->perms |= VSS_READ;
			else if (!strcmp(tmp[num], "VSS_WRITE"))
				u->perms |= VSS_WRITE;
			else if (!strcmp(tmp[num], "AFS_READ"))
				u->perms |= AFS_READ;
			else if (!strcmp(tmp[num], "AFS_WRITE"))
				u->perms |= AFS_WRITE;
			else /* unknown permission */
				PARA_WARNING_LOG("ignoring unknown permission: %s\n",
					tmp[num]);
		}
		para_list_add(&u->node, &user_list);
	}
out:
	if (file_ptr)
		fclose(file_ptr);
	if (ret >= 0)
		return;
	PARA_EMERG_LOG("%s\n", PARA_STRERROR(-ret));
	exit(EXIT_FAILURE);
}

/**
 * initialize the list of users allowed to connecto to para_server
 *
 * \param user_list_file the file containing access information
 *
 * If this function is called a second time, the contents of the
 * previous call are discarded.
 */
void init_user_list(char *user_list_file)
{
	struct user *u, *tmp;
	static int initialized;

	if (initialized) {
		list_for_each_entry_safe(u, tmp, &user_list, node) {
			list_del(&u->node);
			free(u->name);
			rsa_free(u->rsa);
			free(u);
		}
	} else
		INIT_LIST_HEAD(&user_list);
	initialized = 1;
	populate_user_list(user_list_file);
}

/**
 * lookup user in user_list.
 *
 * \param name of the user
 *
 * \return a pointer to the corresponding user struct if the user was found,
 *  \p NULL otherwise.
 */
struct user *lookup_user(const char *name)
{
	struct user *u;
	list_for_each_entry(u, &user_list, node) {
		if (strcmp(u->name, name))
			continue;
		return u;
	}
	return NULL;
}
